# -*- coding:utf-8; tab-width:4; mode:python -*-

from hamcrest import is_, contains_string, all_of
from hamcrest.library.text.stringcontainsinorder import *

from prego import TestCase, Task, File, exists
from prego.shell import Variable

nose = 'nosetests3 -c /dev/null '
prego_cmd = 'bin/prego3 -p -c /dev/null %s'


class LogTests(TestCase):
    def test_show_log_on_wrong_command(self):
        task = Task(desc='wrong')
        cmd = task.command(nose + 'examples/examples.py:Test.test_cmd_wrong_true_and_ls',
                           expected=1)
        strings = [
            "TestFailed: assert that command A.0 expected returncode 0, but was 127",
            ">> begin captured logging <<",
            "[FAIL]   A   Task end - elapsed:"]

        for s in strings:
            task.assert_that(cmd.stderr.content, contains_string(s))

    def test_show_log_on_fail_command(self):
        task = Task(desc='fail')
        cmd = task.command(nose + 'examples/examples.py:Test.test_cmd_false_true',
                           expected=1)
        strings = [
            "TestFailed: assert that command A.0 expected returncode 0, but was 1",
            ">> begin captured logging <<",
            "[FAIL]   A   Task end - elapsed:"]

        for s in strings:
            task.assert_that(cmd.stderr.content, contains_string(s))


class KeepGoingTests(TestCase):
    def test_keep_going_off(self):
        task = Task()
        cmd = task.command(
            prego_cmd % 'examples/examples.py:Test.test_cmd_wrong_true_and_ls',
            expected=1)

        strings = [
            "[ ok ]   A.0 Command 'wrong' code (0:127)",
            "assert that command A.0 expected returncode 0, but was 127",
            "[ -- ]   A.3 Command 'true'",
            "[ -- ]   A.5 assert that command A.3 returncode 0",
            ]

        for s in strings:
            task.assert_that(cmd.stderr.content, contains_string(s))

    def test_keep_going_on(self):
        task = Task()
        cmd = task.command(
            prego_cmd % '--keep-going examples/examples.py:Test.test_cmd_wrong_true_and_ls',
            expected=1)

        strings = [
            "[ ok ]   A.0 Command 'wrong' code (0:127)",
            "assert that command A.0 expected returncode 0, but was 127",
            "[ ok ]   A.3 Command 'true'",
            "[ ok ]   A.5 assert that command A.3 returncode 0",
            "[ ok ]   B.0 Command 'ls'",
            "[ ok ]   B.2 assert that command B.0 returncode 0",
            ]

        for s in strings:
            task.assert_that(cmd.stderr.content, contains_string(s))


class GeneratingCommands(TestCase):
    # atheist/examples/bucle.test
    def test_for(self):
        task = Task()
        for i, retcode in [(0, 2), (1, 0), (2, 2)]:
            task.command('ls -%s' % i, cwd='/tmp', expected=retcode, timeout=None)


class Assertions(TestCase):
    # atheist/examples/conditions.test
    def test_pre_post(self):
        task = Task()
        task.assert_that(Variable('SHELL'), exists())
        task.assert_that(File('/etc/fstab'), exists())
        task.command('cp /etc/fstab /tmp/')
        task.generate_files('/tmp/fstab')

        task.command('ls / > /tmp/kk')  # shell=True)
        task.generate_files('/tmp/kk')

        task.assert_that(File('/tmp/kk').content, all_of(
                contains_string('sbin'),
                contains_string('home'),
                contains_string('boot')))


class OutLoggingTests(TestCase):
    def test_auto_log_stdout(self):
        task = Task()
        cmd = task.command(prego_cmd % '-vo examples/command-outs.py:OK.test_ls_stdout_auto')
        task.assert_that(cmd.stderr.content,
                         contains_string('A.0.out| /etc/passwd'))

    def test_noauto_log_stdout_when_stdout_flag(self):
        task = Task()
        cmd = task.command(prego_cmd % '-vo examples/command-outs.py:OK.test_ls_stdout')
        task.assert_that(cmd.stderr.content,
                         contains_string('A.0.out| /etc/passwd'))

    def test_auto_log_stderr(self):
        task = Task()
        cmd = task.command(prego_cmd % '-ve examples/command-outs.py:OK.test_ls_stderr_auto')
        task.assert_that(cmd.stderr.content,
                         contains_string('A.0.err| /etc/passwd'))

    def test_noauto_log_stderr_when_stderr_flag(self):
        task = Task()
        cmd = task.command(prego_cmd % '-ve examples/command-outs.py:OK.test_ls_stderr')
        task.assert_that(cmd.stderr.content,
                         contains_string('A.0.err| /etc/passwd'))

    def test_vv_implies_nocapture(self):
        task = Task()
        cmd = task.command(prego_cmd % '-vv test/fixtures/output.py:OutputFixture.test_print_text')
        task.assert_that(cmd.stdout.content,
                         contains_string('OutputFixture printed this'))


class AdviceTests(TestCase):
    def test_no_detach_no_timeout(self):
        task = Task(detach=True)
        cmd = task.command(prego_cmd % '-v test/integration/advices.py:Timeout.test_no_detach_no_timeout', expected=None)
        task.assert_that(cmd.stderr.content,
                         contains_string("A.0 No timeout command in a non detached task could block forever!"))
        Task().delay()


class multiline_commands(TestCase):
    def test_escape_linebreaks_on_log(self):
        task = Task()
        task.command(prego_cmd % '-v test/integration/cases.py:multiline_commands.test_writing_out')
        task.assert_that(task.lastcmd.stderr.content,
                         contains_string("hi\\nbye\\nagain"))



# -*- coding: utf-8; mode: python -*-

from unittest import TestCase
import hamcrest

from .unit_tests import AssertionTest

import sys
sys.path.insert(0, '.')

import prego


class FileMatchers(TestCase, AssertionTest):
    def test_file_exists(self):
        c = self.assert_that(prego.File('/etc/fstab'), prego.exists())
        c.eval()

    def test_file_exists_in_cwd(self):
        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('test/$testfilename'), prego.exists())
        prego.commit()

    def test_file_exists_fails(self):
        c = self.assert_that(prego.File('/etc/missing'), prego.exists())

        with self.assertRaises(prego.PregoAssertionFailed):
            c.eval()

    def test_file_do_not_exists(self):
        c = self.assert_that(prego.File('/etc/missing'),
                             hamcrest.is_not(prego.exists()))
        c.eval()

    def test_file_contains(self):
        open('/tmp/prego-content', 'wt').write('this a sample file for prego')
        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('/tmp/prego-content').content,
                      hamcrest.contains_string('sample file'))
        prego.commit()

    def test_non_exising_file_contains(self):
        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('/tmp/kk').content,
                      hamcrest.is_not(hamcrest.contains_string('SOMETHING')))
        prego.commit()

    def test_file_contains_fail(self):
        c = self.assert_that(prego.File('/etc/fstab').content,
                             hamcrest.contains_string('missing'))

        with self.assertRaises(prego.PregoAssertionFailed):
            c.eval()

    def test_file_permissions(self):
        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('/etc/fstab'), prego.has_permissions(0o644))
        t.assert_that(prego.File('/etc/fstab'), prego.has_permissions(0o200))
        t.assert_that(prego.File('/etc/fstab'), prego.has_permissions(0o400))
        t.assert_that(prego.File('/etc/fstab'), prego.has_permissions(0o440))
        t.assert_that(prego.File('/etc/fstab'), prego.has_permissions(0o004))
        prego.commit()

    def test_file_compare(self):
        prego.init()
        a = prego.File('/etc/passwd')
        b = prego.File('/etc/fstab')

        t = prego.Task()
        t.assert_that(a, hamcrest.is_not(b))
        prego.commit()

    def test_file_compare2(self):

        with open('/tmp/a', 'w') as fd:
            fd.write("foobar")

        with open('/tmp/b', 'w') as fd:
            fd.write("foobar")

        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('/tmp/a'), hamcrest.is_(prego.File('/tmp/b')))
        prego.commit()

    def test_file_compare_fail(self):
        prego.init()
        t = prego.Task()
        t.assert_that(prego.File('/etc/fstab'), hamcrest.is_not(prego.File('/etc/passwd')))
        prego.commit()



# -*- coding:utf-8; tab-width:4; mode:python -*-

import sys
import logging
import unittest
import contextlib
from commodity.path import child_relpath

from .runner import init, Runner
from .exc import TestFailed
from .tools import StatusFilter
from .const import Status, term
from . import gvars


class PregoTestCase(object):
    def __init__(self, testcase, methodname, testpath):
        self.testcase = testcase
        self.methodname = methodname
        self.status = Status.NOEXEC

        self.name = "%s:%s.%s" % (child_relpath(testpath), testcase.__class__.__name__, methodname)
        self.log = logging.getLogger(self.name)
        self.log.setLevel(logging.INFO)
        self.log.addFilter(StatusFilter(self))
        init()

    def commit(self):
        self.status = Status.UNKNOWN
        self.log.info(Status.indent('=') + term().reverse(' INI ') + ' $name')
        try:
            Runner(gvars.tasks).run()
            self.status = Status.OK
        except TestFailed as test_failed:
            self.status = Status.FAIL
            raise test_failed  # shrink traceback
        except Exception:
            self.status = Status.ERROR
            raise
        finally:
            self.log.info('$status ' + term().reverse(' END ') + ' $name')
            init()


class TestCase(unittest.TestCase):
    def _callSetUp(self, testMethod=None):
        if testMethod:
            gvars.testpath = testpath = testMethod.__code__.co_filename
            self.prego_case = PregoTestCase(self, self._testMethodName, testpath)
        self.setUp()

    def _callTestMethod(self, method):
        method()
        self.prego_case.commit()

    def _callTearDown(self):
        self.tearDown()

    def run(self, result=None):
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, 'startTestRun', None)
            if startTestRun is not None:
                startTestRun()

        result.startTest(self)

        testMethod = getattr(self, self._testMethodName)
        if (getattr(self.__class__, "__unittest_skip__", False) or
            getattr(testMethod, "__unittest_skip__", False)):
            # If the class or method was skipped.
            try:
                skip_why = (getattr(self.__class__, '__unittest_skip_why__', '')
                            or getattr(testMethod, '__unittest_skip_why__', ''))
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return
        expecting_failure_method = getattr(testMethod,
                                           "__unittest_expecting_failure__", False)
        expecting_failure_class = getattr(self,
                                          "__unittest_expecting_failure__", False)
        expecting_failure = expecting_failure_class or expecting_failure_method
        outcome = _Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self):
                self._callSetUp(testMethod)
            if outcome.success:
                outcome.expecting_failure = expecting_failure
                with outcome.testPartExecutor(self, isTest=True):
                    self._callTestMethod(testMethod)
                outcome.expecting_failure = False
                with outcome.testPartExecutor(self):
                    self._callTearDown()

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, 'stopTestRun', None)
                if stopTestRun is not None:
                    stopTestRun()

            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome -> outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None

            # clear the outcome, no more needed
            self._outcome = None


# patched unittest.case._Outcome
class _Outcome(object):
    def __init__(self, result=None):
        self.expecting_failure = False
        self.result = result
        self.result_supports_subtests = hasattr(result, "addSubTest")
        self.success = True
        self.skipped = []
        self.expectedFailure = None
        self.errors = []

    @contextlib.contextmanager
    def testPartExecutor(self, test_case, isTest=False):
        old_success = self.success
        self.success = True
        try:
            yield
        except KeyboardInterrupt:
            raise
        except unittest.case.SkipTest as e:
            self.success = False
            self.skipped.append((test_case, str(e)))
        except unittest.case._ShouldStop:
            pass
        except:
            exc_info = list(sys.exc_info())
            if exc_info[0] == TestFailed:
                exc_info[2] = None  # remove traceback

            if self.expecting_failure:
                self.expectedFailure = exc_info
            else:
                self.success = False
                self.errors.append((test_case, exc_info))
            # explicitly break a reference cycle:
            # exc_info -> frame -> exc_info
            exc_info = None
        else:
            if self.result_supports_subtests and self.success:
                self.errors.append((test_case, None))
        finally:
            self.success = self.success and old_success



# -*- coding:utf-8; tab-width:4; mode:python -*-

import os
import filecmp

from commodity.type_ import checked_type
from commodity.str_ import Printable

from .tools import file_types
from .tools import Interpolator


class DeferredItem(object):
    def __init__(self):
        self.task = None

    def config(self, task):
        self.task = task


class DeferredAttr(DeferredItem):
    "An object attribute delegated for late binding"

    def __init__(self, obj, name, attr):
        self.obj = obj
        self.name = name
        self.attr = attr

    def resolve(self):
        return getattr(self.obj, self.attr)

    def __str__(self):
        return 'deferred %s.%s' % (self.obj.__class__.__name__, self.name)


class DeferredContent(DeferredItem):
    def __init__(self, fd):
        super(DeferredContent, self).__init__()
        self.fd = fd

    def resolve(self):
        try:
            return self.fd.read()
        except IOError:
            return None

    def __str__(self):
        return "{0} has".format(self.fd)


class File(Printable):
    def __init__(self, path, fd=None):
        self.path = Interpolator().apply(checked_type(str, path))
        self.fd = fd

    def __eq__(self, other):
        return filecmp.cmp(self.path, other.path)

    @classmethod
    def from_fd(cls, fd):
        "Creates prego.File from standard python file descriptor"
        fd = checked_type(file_types, fd)
        assert not fd.closed, fd
        return File(fd.name, fd)

    def read(self):
        with open(self.path) as fd:
            return fd.read()

    def readline(self):
        with open(self.path) as fd:
            return fd.readline()

    def write(self, data):
        self._assure_open()
        self.fd.write(data)

    def flush(self):
        self._assure_open()
        self.fd.flush()

    @property
    def closed(self):
        return self.fd.closed

    def close(self):
        assert not self.fd.closed
        self.fd.close()

    def _assure_open(self):
        if self.fd is None:
            self.fd = file(self.path, 'w', 0)

        if self.fd.closed:
            raise ValueError('%s was closed' % self.fd.name)

    def find(self, substring):
        raise TypeError("use 'content' attribute to refer file content")

    def exists(self):
        return os.path.exists(self.path)

    def remove(self):
        if self.exists():
            os.remove(self.path)

    @property
    def content(self):
        return DeferredContent(self)

    def __str__(self):
        return "File {0!r}".format(self.path)



# -*- coding:utf-8; tab-width:4; mode:python -*-

import os
from signal import SIGTERM
import time
from functools import partial
import io

import hamcrest

from commodity.type_ import checked_type
from commodity.os_ import SubProcess, FileTee, FuncAsTextFile
from commodity.log import PrefixLogger

from .assertion import Assertion, Matcher
from .tools import create_file, create_logger, Interpolator, file_types
from .const import PREGO_TMP, AUTO, INDENTST
from .item import File, DeferredAttr
from . import config
from . import gvars
from .const import Status, term

NoneType = type(None)


DEFAULT = '__DEFAULT__'


class Command(Assertion):
    def __init__(self, task, cmdline,
                 stdout=None, stderr=None,
                 expected=DEFAULT, timeout=DEFAULT,
                 signal=None, cwd=None, env={}):

        self._returncode = None
        self.interpolator = Interpolator()
        self.cmdline = self.interpolator.apply(checked_type(str, cmdline))

        self._set_expected(expected)
        super(Command, self).__init__(task)

        self._stdout = self.interpolator.apply(stdout)
        self._stderr = self.interpolator.apply(stderr)

        self._set_cwd(cwd)

        self.tinit = self.tend = None

        self._set_timeout(timeout)
        self._set_signal(signal)
        self._set_env(env)

        self.sp = None
        self._enable_outs_if_requested()

        if self.timeout is None and not self.task.detach:
            self.log.warning('%s$name %sNo timeout command in a non detached task could block forever!%s', INDENTST, term().bold_red, term().normal)

    def _set_cwd(self, cwd):
        cwd = cwd or gvars.context.get('cwd', '.')
        self.cwd = self.interpolator.apply(checked_type(str, cwd))

    def _set_expected(self, expected):
        if expected == DEFAULT:
            expected = gvars.context.get('expected', 0)
        self.expected = checked_type((int, NoneType), expected)

    def _set_timeout(self, timeout):
        if timeout == DEFAULT:
            timeout = gvars.context.get('timeout', 5)
        self.timeout = checked_type((int, NoneType), timeout)

    def _set_signal(self, signal):
        signum = signal or gvars.context.get('signal', SIGTERM)
        self.signal = checked_type(int, signum)

    def _set_env(self, env):
        self.env = os.environ.copy()
        context_env = checked_type(dict, gvars.context.get('env', {}))
        self.env.update(context_env)
        self.env.update((str(k), str(v)) for k, v in env.items())

    def do_eval(self):
        return self.run()

    def is_running(self):
        return self.status is Status.UNKNOWN and not self.terminated

    @property
    def terminated(self):
        if self.sp is None:
            return False

        return self.sp.terminated()

    def _enable_outs_if_requested(self):
        self._enable_out('out', bool(self._stdout) or config.stdout)
        self._enable_out('err', bool(self._stderr) or config.stderr)

    def _enable_out(self, outname, state=True):
        assert outname in ['out', 'err'], "Internal Error"
        if not state:
            return

        out = getattr(self, "_std%s" % outname)
        if out is None and state:
            out = AUTO

        if out is AUTO:
            out = os.path.join(PREGO_TMP, '%s.%s' % (self.name, outname))

        if isinstance(out, str):
            out = create_file(out, 'wb', 0)

        if isinstance(out, file_types):
            assert out.writable(), "%s must be writable" % out.name
            out = File.from_fd(out)

        setattr(self, "_std%s" % outname, out)
        self.task.gen.append(out)

    @property
    def stdout(self):
        if self._stdout is None:
            self._stdout = AUTO
            self._enable_out('out')

        return self._stdout

    @property
    def stderr(self):
        if self._stderr is None:
            self._stderr = AUTO
            self._enable_out('err')

        return self._stderr

    def _any_out_activated(self):
        return any([self.stdout, self.stderr])

    def run(self):
        def decorate_out_with_logger(out, suffix, activate):
            if out is None:
                return

            if not activate:
                return out

            out_logger_name = '{0}.{1}'.format(self.name, suffix)
            out_logger = create_logger(out_logger_name)
            prefix = '%s%s| ' % (9 * ' ', out_logger.name)
            return FileTee(
                out, to_textfile_wrapper(
                    FuncAsTextFile(out_logger.info, prefix)))

        stdout = decorate_out_with_logger(self._stdout, 'out', config.stdout)
        stderr = decorate_out_with_logger(self._stderr, 'err', config.stderr)

        if self._any_out_activated():
            self.log_status()

        logger = PrefixLogger(self.log, '%s%s ' % (INDENTST, self.name))

        self.tinit = time.time()
        self.sp = SubProcess(
            self.cmdline,
            stdout=stdout, stderr=stderr,
            close_outs=True,
            cwd=self.cwd,
            shell=True,
            signal=self.signal,
            env=self.env,
            logger=logger)

        if not self.terminated:
            self.log_status()

        self.wait()
        self.tend = time.time()
        return True

    @property
    def returncode(self):
        if self._returncode is not None:
            return self._returncode

        # FIXME: is this required?
        return DeferredAttr(self, 'returncode', 'sync_returncode')

    @property
    def sync_returncode(self):
        return self.sp.returncode

    def _current_elapsed(self):
        return time.time() - self.tinit

    def elapsed(self):
        return self.tend - self.tinit

    def on_time(self):
        return self._current_elapsed() < self.timeout

    def wait(self):
        def terminate():
            if self.terminated:
                return

            msg = '{0}$name timeout exceeded: {1:1.2f}s >= {2:1.2f}s, terminating...'
            self.log.debug(msg.format(INDENTST, self._current_elapsed(),
                                      self.timeout))
            self.terminate()

        if self.timeout is not None:
            while self.on_time() and not self.terminated:
                time.sleep(0.05)

            terminate()

        self.sp.wait()
        self._returncode = self.sp.returncode

    def terminate(self):
        if self.sp is None:
            self.log.warning('%s$name terminate: task never started' % INDENTST)
            return

        self.sp.terminate()
        assert self.terminated

    def __unicode__(self):
#        cmd = self.sp or repr(self.cmdline)
        expected = self.expected
        if expected is None:
            expected = ''

        returncode = self.returncode
        if isinstance(returncode, DeferredAttr):
            returncode = ''

        timeout = self.timeout if self.timeout is not None else ''

        elapsed = ''
        if self.terminated and self.tend is not None:
            elapsed = '%.2f' % (self.tend - self.tinit)

        time = ''
        if timeout or elapsed:
            time = ' time {0}:{1}'.format(timeout, elapsed)

        cwd = ''
#        if self.cwd != self.interpolator.testdir:
#            cwd = ' cwd:{0}'.format(self.cwd)

        return u"Command {0!r} code ({1}:{2}){3}{4}".format(
            self.cmdline, expected, returncode, time, cwd)

    def brief(self):
        return "command " + self.name


class CommandExitsWith(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def _matches(self, cmd):
        self.cmd = checked_type(Command, cmd)
        return self.cmd.returncode == self.expected

    def describe_to(self, description):
        description.append_text("returncode %s" % (self.expected))

    def describe_mismatch(self, cmd, mismatch_description):
        mismatch_description.append_text('was %s' % self.cmd.returncode)

    def brief(self):
        return u'exit code'


def exits_with(expected):
    return CommandExitsWith(expected)


def killed_by(signal):
    return CommandExitsWith(-signal)


class RanForTime(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def _matches(self, cmd):
        self.cmd = checked_type(Command, cmd)
        return hamcrest.is_(self.expected).matches(cmd.elapsed())

    def describe_to(self, description):
        description.append_text("execution time %ss" % (self.expected))

    def describe_mismatch(self, cmd, mismatch_description):
        mismatch_description.append_text('was %.2fs' % self.cmd.elapsed())


def ran_for_time(secs):
    return RanForTime(secs)


class to_textfile_wrapper:
    def __init__(self, fd):
        self.fd = fd

    def write(self, data):
        self.fd.write(data.decode(errors='ignore'))

    def flush(self):
        pass

    @property
    def closed(self):
        return self.fd.closed

    def close(self):
        pass



# -*- coding:utf-8; tab-width:4; mode:python -*-

import os
import prego
import hamcrest
import pwd

TESTDIR = os.path.abspath(os.path.split(__file__)[0])
BASEDIR = os.path.dirname(TESTDIR)
PID = os.getpid()
LOGIN = pwd.getpwuid(os.getuid())[0]


class CommandInterpolation(prego.TestCase):
    def echo(self, var):
        task = prego.Task(var)
        cmd = task.command('echo ' + var, stdout=prego.AUTO)
        return task, cmd

    def test_basedir(self):
        task, cmd = self.echo('$basedir')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace('.'))

    def test_fullbasedir(self):
        task, cmd = self.echo('$fullbasedir')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(BASEDIR))

    def test_testdir(self):
        task, cmd = self.echo('$testdir')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace('test'))

    def test_fulltestdir(self):
        task, cmd = self.echo('$fulltestdir')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(TESTDIR))

    def test_testfilename(self):
        task, cmd = self.echo('$testfilename')
        expected = os.path.split(__file__)[1].rstrip('c')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(expected))

    def test_pid(self):
        task, cmd = self.echo('$pid')
        task.assert_that(cmd.stdout.content, hamcrest.contains_string(str(PID)))

    def test_tmp_base(self):
        expected = '/tmp/prego-' + LOGIN
        task, cmd = self.echo('$tmpbase')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(expected))

    def test_tmp(self):
        expected = '/tmp/prego-{0}/{1}'.format(LOGIN, PID)
        task, cmd = self.echo('$tmp')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(expected))


class KeywordsInterpolation(prego.TestCase):
    def test_cwd(self):
        task = prego.Task()
        cmd = task.command('pwd', cwd='$testdir')
        task.assert_that(cmd.stdout.content,
                         hamcrest.equal_to_ignoring_whitespace(TESTDIR))

    def test_stdout(self):
        task = prego.Task()
        task.command('echo hi', stdout='$basedir/task.out')
        task.assert_that(prego.File('$basedir/task.out'),
                         prego.exists())



# -*- coding:utf-8; tab-width:4; mode:python -*-

import os
import pwd
import string
import site
import sys
from pprint import pprint

import blessings
from commodity.pattern import memoized
from commodity.os_ import resolve_path

from . import config


def get_site_base(path):
    if not path or 'lib' not in path:
        return None

    return path.split('lib')[0]


AUTO = '__AUTO__'
PREGO_TMP_BASE = os.path.join('/tmp', 'prego-{0}'.format(pwd.getpwuid(os.getuid())[0]))
PREGO_TMP = os.path.join(PREGO_TMP_BASE, str(os.getpid()))
INDENTST = 9 * ' '
IDENTIFIERS = string.ascii_uppercase + string.ascii_lowercase

install_paths = [
    '/usr',          # debian (sys.prefix)
    '/'              # virtualenv
    '/usr/local',    # pip
    site.USER_BASE]  # pip

site_base = get_site_base(os.environ.get('PYTHONPATH'))
if site_base:
    install_paths.append(site_base)

data_files_suffix = 'share/lib/prego3'
config_path = ['.'] + [os.path.join(p, data_files_suffix) for p in install_paths]

# print("--config_path--")
# pprint(config_path)
# print("--resolve_path--")
# pprint(resolve_path('config.spec', config_path))
# print(__file__)
# print(site.USER_BASE)

PREGO_SPECS = os.path.abspath(resolve_path('config.spec', config_path)[0])
PREGO_CMD_DEFAULTS = os.path.abspath(resolve_path('defaults.config', config_path)[0])
USER_CONFIG = os.path.abspath(os.path.join(os.environ['HOME'], '.prego'))
CWD_CONFIG = os.path.join(os.getcwd(), '.prego')


@memoized
def term():
    return blessings.Terminal(force_styling=config.force_color)


class Status:
    _FAIL     = 0
    _OK       = 1
    _ERROR    = 2
    _NOEXEC   = 3
    _UNKNOWN  = 4
    _SOFTFAIL = 5
    _SOFTOK   = 6

    stringfied_values = {
        _FAIL:     ('FAIL',    'FAIL'),
        _OK:       ('OK',      'OK'),
        _ERROR:    ('ERROR',   '!!'),
        _NOEXEC:   ('NOEXEC',  '--'),
        _UNKNOWN:  ('UNKNOWN', '??'),
        _SOFTFAIL: ('fail',    'fail'),
        _SOFTOK:   ('ok',      'ok'),
        }

    color_map = {
        _FAIL:     lambda: term().bold_red,
        _OK:       lambda: term().bold_green,
        _NOEXEC:   lambda: term().normal,
        _ERROR:    lambda: term().bold_red,
        _UNKNOWN:  lambda: term().normal,
        _SOFTFAIL: lambda: term().red,
        _SOFTOK:   lambda: term().green,
        }

    def __init__(self, value=None):
        if value is None:
            value = self._NOEXEC
        self.value = value

    def __eq__(self, other):
        if all(x.value in [self.FAIL.value, self.SOFTFAIL.value] for x in [self, other]):
            return True

        if all(x.value in [self.OK.value, self.SOFTOK.value] for x in [self, other]):
            return True

        return self.value == other.value

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.stringfied_values[self.value][0]

    def is_bad(self):
        return self.value in [self._FAIL, self._ERROR]

    def pretty(self):
        retval = "{0:^4}".format(self.stringfied_values[self.value][1])
        retval = str.join('', [self.color_map[self.value]()] + [retval, term().normal])
        return "[{0}]".format(retval)

    @classmethod
    def define_statuses(cls):
        Status.FAIL     = Status(Status._FAIL)
        Status.OK       = Status(Status._OK)
        Status.ERROR    = Status(Status._ERROR)
        Status.NOEXEC   = Status(Status._NOEXEC)
        Status.UNKNOWN  = Status(Status._UNKNOWN)
        Status.SOFTFAIL = Status(Status._SOFTFAIL)
        Status.SOFTOK   = Status(Status._SOFTOK)

    def log(self, logger, msg):
        if self in [Status.NOEXEC, Status.UNKNOWN]:
            func = logger.debug
        elif self.is_bad():
            func = logger.error
        else:
            func = logger.info

        func(msg)

    @classmethod
    def indent(cls, char=' '):
        return (6 * char) + ' '


Status.define_statuses()



# -*- coding:utf-8; tab-width:4; mode:python -*-

import sys
import os
import logging
import threading
import argparse

from logging import StreamHandler
from logging import Handler
log = logging.getLogger('nose.plugins.prego')
log.addHandler(logging.StreamHandler())

import nose
from nose.plugins import logcapture
from nose.plugins.logcapture import FilterSet

from commodity.type_ import module_to_dict
from commodity.log import CapitalLoggingFormatter, NullHandler
from commodity.pattern import MetaBunch
from commodity.args import parser, add_argument, args

from . import config
from .tools import set_logger_default_formatter, StatusFilter, update_obj
from . import const

logging.getLogger('nose.plugins.manager').addHandler(NullHandler())
plugins_logger = logging.getLogger('nose')
plugins_logger.propagate = False

rdflog = logging.getLogger('rdflib')
# rdflog.propagate = False
rdflog.addFilter(StatusFilter())
rdflog.addHandler(NullHandler())

logging.getLogger().addFilter(StatusFilter())


class MyMemoryHandler(Handler):
    def __init__(self, logformat, logdatefmt, filters):
        Handler.__init__(self)
        fmt = CapitalLoggingFormatter(logformat, logdatefmt)
        self.setFormatter(fmt)
        self.filterset = FilterSet(filters)
        self.buffer = []
    def emit(self, record):
        self.buffer.append(self.format(record))
    def flush(self):
        pass # do nothing
    def truncate(self):
        self.buffer = []
    def filter(self, record):
        return self.filterset.allow(record.name)
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['lock']
        return state
    def __setstate__(self, state):
        self.__dict__.update(state)
        self.lock = threading.RLock()



# if isinstance(logcapture.MyMemoryHandler, BufferingHandler):
logcapture.MyMemoryHandler = MyMemoryHandler
logcapture.LogCapture.logformat = '%(levelcapital)s. %(message)s'
logcapture.clear = True


def run():
    parser.prog = 'prego'

    # behaviour
    parser.add_argument('-c', '--config', metavar='FILE',
                        help='explicit config file')
    parser.add_argument('-k', '--keep-going', action='store_true',
                        help="continue even with failed assertion or tests")
    parser.add_argument('-d', '--dirty', action='store_true',
                        help="do not remove generated files")

    # output
    parser.add_argument('-o', '--stdout', action='store_true',
                        help='print tests stdout')
    parser.add_argument('-e', '--stderr', action='store_true',
                        help='print tests stderr')
    parser.add_argument('-t', '--time-tag', dest='timetag', action='store_true',
                        help='Include time info in logs')

    parser.add_argument('-v', '--verbose', dest='verbosity', action='count',
                        help='increase log verbosity')

    # styling
    parser.add_argument('-p', '--plain', action='store_false', dest='color',
                        help='avoid colors and styling in output')
    parser.add_argument('-f', '--force-color', action='store_true', dest='force_color',
                        help='force colors and styling in output')

    # pass throw nose options
    parser.add_argument('nose', metavar='nose-args', nargs=argparse.REMAINDER)

    parser.load_config_file(const.PREGO_CMD_DEFAULTS)
    parser.parse_args()

    for x in range(args.nose.count('--')):
        args.nose.remove('--')

    if args.config:
        parser.load_config_file(os.path.abspath(args.config))
    else:
        parser.load_config_file(const.USER_CONFIG)
        parser.load_config_file(const.CWD_CONFIG)

    update_obj(config, args)

    if config.timetag:
        logcapture.LogCapture.logformat = '%(asctime)s [%(levelcapital)s] %(message)s'

    if config.verbosity:
        args.nose += ['--nologcapture', '--quiet']

        if config.verbosity == 1:
            loglevel = logging.INFO
        elif config.verbosity >= 2:
            loglevel = logging.DEBUG
            args.nose += ['--nocapture']

        if config.verbosity > 2:
            config.stderr = config.stdout = True

        root = logging.getLogger()
        root.setLevel(loglevel)

        handler = StreamHandler()
        root.addHandler(handler)
        set_logger_default_formatter(root)

    if config.force_color:
        logging.warning("Option -f/--force-color is deprecated")
        sys.exit(1)

    nose.main(argv=['dummy'] + args.nose)



# -*- mode: python; coding: utf-8 -*-
import time

import hamcrest
from hamcrest.core.base_matcher import BaseMatcher

from commodity.thread_ import start_new_thread
from commodity.type_ import checked_type
from commodity.str_ import Printable
from commodity.log import UniqueFilter
from commodity.pattern import memoizedproperty
from commodity.log import CallerData
from commodity.path import child_relpath

from .const import Status, INDENTST, term, IDENTIFIERS, PREGO_TMP
from .exc import (PregoAssertionException, log_traceback, UserBreak)
from .tools import Interpolator
from .item import File
from .assertion import (Assertion, DeferredAssertion, PollDeferredAssertion, Delay)
from .command import Command, exits_with, ran_for_time
from . import config
from . import gvars
from .assertion import Matcher

from .tools import create_logger, StatusFilter
# set_logger_default_formatter()


class MatcherRequiredError(Exception):
    pass


class Task(Printable):
    identifiers = []

    def __init__(self, desc='', detach=False):

        # print "stdout:", config.stdout
        # print "stderr:", config.stderr
        # print "keep-going:", config.keep_going

        gvars.tasks.append(self)
        self.interpolator = Interpolator()
        self.desc = checked_type(str, desc)
        self.detach = checked_type(bool, detach)

        self.gen = []
        self.thread = None
        self.status = Status.NOEXEC
        self.tinit = None
        self.elapsed = 0
        self.reason = None

        self.assertions = []

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    @classmethod
    def get_index(cls, objid):
        try:
            i = cls.identifiers.index(objid)
            return IDENTIFIERS[i]
        except IndexError:
            return str(i - 52)
        except ValueError:
            cls.identifiers.append(objid)
            return cls.get_index(objid)

    @property
    def index(self):
        return self.get_index(id(self))

    @property
    def name(self):
        return self.index

    def gen_assertion_index(self):
        return len(self.assertions)

    @memoizedproperty
    def log(self):
        retval = create_logger(self.name)
        retval.addFilter(StatusFilter(self))
        retval.addFilter(UniqueFilter())
        return retval

    def terminate(self):
        self.log.debug("%s$name   waiting for detached task", INDENTST)

        for a in self.assertions:
            a.terminate()

    def run(self):
        if self.detach:
            self.detached_run()
            return

        self.sync_run()

    def detached_run(self):
        self.thread = start_new_thread(self.sync_run)
        self.wait_starts_or_finish()

    def current_elapsed(self):
        return time.time() - self.tinit

    def wait_starts_or_finish(self):
        time.sleep(0.1)

        while 1:
            if self.is_finished() or self.is_running():
                break

            time.sleep(0.1)

            if self.current_elapsed() > 3:
                self.log.critical("%s$name does not start after 3s", INDENTST)
                break

    def sync_run(self):
        self.task_start()

        assertion_iter = iter(self.assertions)
        try:
            for a in assertion_iter:
                a.eval()
                a.log_status()

            self.status = Status.OK

        except PregoAssertionException as e:
            self.set_reason(Status.FAIL, e.assertion)

        except KeyboardInterrupt as e:
            a.status = Status.ERROR
            a.log_status('USER BREAK')
            a.terminate()
            self.set_reason(Status.ERROR, a)
            raise UserBreak()

        except Exception as e:
            # FIXME: convert e to an Assertion
            # self.set_reason(Status.ERROR, e)
            log_traceback(self.log, prefix='%s| ' % INDENTST)

        finally:
            for a in assertion_iter:
                try:
                    if config.keep_going:
                        a.eval()
                    a.log_status()
                except PregoAssertionException:
                    pass

            elapsed = self.current_elapsed()
            self.log.info("$status   $name   Task end - elapsed: %.2fs", elapsed)

    def task_start(self):
        self.tinit = time.time()
        detach_advice = ' [DETACHED]' if self.detach else ''
        msg = '$status   %s$name   Task starts%s: %s%s' % (
            term().bold, detach_advice, self.desc, term().normal)
        self.log.debug(msg)
        self.status = Status.UNKNOWN

    def assert_that(self, actual, matcher=None, reason=''):
        if matcher and not isinstance(matcher, BaseMatcher):
            raise MatcherRequiredError("%s should be a hamcrest Matcher" % str(matcher))
        assertion = DeferredAssertion(self, actual, matcher, reason)
        assertion.caller = CallerData()
        return assertion

    def wait_that(self, actual, matcher, reason='', delta=1, timeout=5):
        assertion = PollDeferredAssertion(
            self, actual, matcher, reason, delta, timeout)
        assertion.caller = CallerData()
        return assertion

#    def fail(self):
#        return FailAssertion(self)

    def command(self, cmdline, **kargs):
        cmd = Command(self, cmdline, **kargs)
        cmd.caller = CallerData()

        if cmd.timeout is not None:
            assertion = DeferredAssertion(
                self, cmd, ran_for_time(hamcrest.less_than(cmd.timeout)))
            assertion.caller = CallerData()

        if cmd.expected is not None:
            assertion = DeferredAssertion(self, cmd, exits_with(cmd.expected))
            assertion.caller = CallerData()

        return cmd

    def delay(self, n=1):
        return Delay(self, n)

    @property
    def lastcmd(self):
        try:
            return [x for x in self.assertions if isinstance(x, Command)][-1]
        except IndexError:
            raise IndexError("No command assertions was defined")

    def is_running(self):
        return self.status == Status.UNKNOWN

    def is_finished(self):
        return self.status not in [Status.UNKNOWN, Status.NOEXEC]

    def wait_detached(self, timeout=None):
        if self.thread:
            self.thread.join(timeout)

    def set_reason(self, status, reason):
        self.status = status
        self.reason = checked_type(Assertion, reason)
        reason.log_status()

    def generate_files(self, *fnames):
        self.gen.extend([File(x) for x in fnames])

    def remove_gen(self):
        def get_relpaths(files):
            return [child_relpath(f.path, start=PREGO_TMP) for f in files]

        if not self.gen:
            return

        if config.dirty:
            self.log.info('%s  $name dirty mode: generated files not removed: %s ',
                          Status.indent(), get_relpaths(self.gen))
            return

        self.log.debug("%s$name   removing %s", INDENTST, get_relpaths(self.gen))

        for f in self.gen:
            f.remove()

    # FIXME: may have serveral command (or none)
    # FIXME: log format
    def __unicode__(self):
        return self.index


class Running(Matcher):
    def _matches(self, cmd):
        return cmd.is_running()

    def describe_to(self, description):
        description.append_text('is running ')


def running():
    return Running()


def terminated():
    return hamcrest.is_not(Running())


# FIXME
def command(cmd, **kargs):
    test = Task()
    test.command(cmd, **kargs)
    return test



# -*- coding:utf-8; tab-width:4; mode:python -*-
'''
Debian related assertion
'''

from six import BytesIO

from commodity.type_ import checked_type
from commodity.os_ import SubProcess
from commodity.str_ import Printable

from .assertion import Matcher


class Package(Printable):
    def __init__(self, name):
        self.name = checked_type(str, name)

    def __str__(self):
        return self.name


class DebPackageInstalled(Matcher):
    def __init__(self, min_version):
        self.version = min_version
        super(DebPackageInstalled, self).__init__()

    def _matches(self, package):
        self.package = package

        out = BytesIO()
        sp = SubProcess('dpkg -l %s | grep ^ii' % self.package.name,
                        stdout=out, shell=True)
        retval = not sp.wait()

        if retval and self.version is not None:
            installed = out.getvalue().decode(errors='ignore').split()[2].strip()
            retval &= installed >= self.version

            print("%s: pkg:%s inst:%s req:%s" %
                  (self.__class__.__name__, self.package,
                   installed, self.version))

        return retval

    def describe_to(self, description):
        description.append_text('package is installed')


def installed(min_version=None):
    return DebPackageInstalled(min_version)



# -*- mode:python; coding:utf-8; tab-width:4 -*-

from prego import TestCase, Task
from prego.debian import Package, installed


class pip_tests(TestCase):
    def test_install(self):
        task = Task()
        task.assert_that(Package('python3-virtualenv'), installed())
        task.assert_that(Package('virtualenv'), installed())
        task.command('rm dist/*', expected=None)
        task.command('python setup.py sdist', timeout=20)
        task.command('virtualenv --clear venv', timeout=10)
        task.command('. venv/bin/activate; pip install dist/prego*; echo y | pip uninstall prego; deactivate', timeout=50)



# -*- coding:utf-8; tab-width:4; mode:python -*-

import logging
import traceback

#from .tools import Printable


class TestResultException(AssertionError):
    def __init__(self, task):
        super(Exception, self).__init__(
            "%s\n%s" % (task.reason, task.reason.caller))


class TestFailed(TestResultException):
    pass


class TestError(TestResultException):
    pass


class UserBreak(Exception):
    pass


class ConfigError(Exception):
    pass


class PregoAssertionException(Exception):
    def __init__(self, assertion):
        self.assertion = assertion
        super(Exception, self).__init__(assertion)


class PregoAssertionFailed(PregoAssertionException):
    pass


class PregoAssertionError(PregoAssertionException):
    pass


class CommandAlreadyDefined(Exception):
    pass


def log_exc(exc, logger=logging):
    for line in str(exc).splitlines():
        if line.strip():
            logger.error('| ' + line)


def log_traceback(logger=logging, prefix='| '):
    for line in traceback.format_exc().splitlines():
        if line.strip():
            logger.error(prefix + line)



#!/usr/bin/python3

# Copyright (C) 2012-2018 David Villa Alises
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from distutils.core import setup
from setuptools import find_packages


def get_reqs():
    with open('requirements.txt') as fd:
        return fd.readlines()


exec(open('version.py').read())

config = dict(
    name             = 'prego3',
    version          = __version__,
    description      = 'System test framework for POSIX shells (python3 version)',
    author           = 'David Villa Alises',
    author_email     = 'David.Villa@gmail.com',
    entry_points     = {
        'console_scripts': [ 'prego3=prego:run' ]
    },
    data_files       = [
        ('share/doc/prego3', ['README.md']),
        ('share/lib/prego3', ['config.spec', 'defaults.config']),
    ],
    url              = 'https://github.com/davidvilla/prego',
    license          = 'GPLv3',
    packages         = find_packages(),
    provides         = ['prego3'],
    long_description_content_type="text/markdown",
    long_description = open('README.md').read(),
    install_requires = get_reqs(),
    classifiers      = [
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python :: 3',
    ],
)

setup(**config)
